﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Selection;

namespace HeuristicLab.Algorithms.RAPGA {
  /// <summary>
  /// An operator which represents the main loop of a relevant alleles preserving genetic algorithm.
  /// </summary>
  [Item("RAPGAMainLoop", "An operator which represents the main loop of a relevant alleles preserving genetic algorithm.")]
  [StorableClass]
  public sealed class RAPGAMainLoop : AlgorithmOperator {
    #region Parameter properties
    public ValueLookupParameter<IRandom> RandomParameter {
      get { return (ValueLookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (ValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ValueLookupParameter<IOperator> SelectorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Selector"]; }
    }
    public ValueLookupParameter<IOperator> CrossoverParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Crossover"]; }
    }
    public ValueLookupParameter<PercentValue> MutationProbabilityParameter {
      get { return (ValueLookupParameter<PercentValue>)Parameters["MutationProbability"]; }
    }
    public ValueLookupParameter<IOperator> MutatorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Mutator"]; }
    }
    public ValueLookupParameter<IOperator> EvaluatorParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Evaluator"]; }
    }
    public ValueLookupParameter<IntValue> ElitesParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["Elites"]; }
    }
    public ValueLookupParameter<IntValue> MaximumGenerationsParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["MaximumGenerations"]; }
    }
    public ValueLookupParameter<VariableCollection> ResultsParameter {
      get { return (ValueLookupParameter<VariableCollection>)Parameters["Results"]; }
    }
    public ValueLookupParameter<IOperator> AnalyzerParameter {
      get { return (ValueLookupParameter<IOperator>)Parameters["Analyzer"]; }
    }
    public ValueLookupParameter<IntValue> EvaluatedSolutionsParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["EvaluatedSolutions"]; }
    }
    public ValueLookupParameter<IntValue> PopulationSizeParameter {
      get { return (ValueLookupParameter<IntValue>)Parameters["PopulationSize"]; }
    }
    public IValueLookupParameter<IntValue> MinimumPopulationSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MinimumPopulationSize"]; }
    }
    public IValueLookupParameter<IntValue> MaximumPopulationSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumPopulationSize"]; }
    }
    public IValueLookupParameter<DoubleValue> ComparisonFactorParameter {
      get { return (IValueLookupParameter<DoubleValue>)Parameters["ComparisonFactor"]; }
    }
    public IValueLookupParameter<IntValue> EffortParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["Effort"]; }
    }
    private ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }

    public IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private RAPGAMainLoop(bool deserializing) : base(deserializing) { }
    private RAPGAMainLoop(RAPGAMainLoop original, Cloner cloner) : base(original, cloner) { }
    public RAPGAMainLoop()
      : base() {
      Initialize();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new RAPGAMainLoop(this, cloner);
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new ValueLookupParameter<IRandom>("Random", "A pseudo random number generator."));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization", "True if the problem is a maximization problem, otherwise false."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "The value which represents the quality of a solution."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Selector", "The operator used to select solutions for reproduction."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Crossover", "The operator used to cross solutions."));
      Parameters.Add(new ValueLookupParameter<PercentValue>("MutationProbability", "The probability that the mutation operator is applied on a solution."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Mutator", "The operator used to mutate solutions."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Evaluator", "The operator used to evaluate solutions. This operator is executed in parallel, if an engine is used which supports parallelization."));
      Parameters.Add(new ValueLookupParameter<IntValue>("Elites", "The numer of elite solutions which are kept in each generation."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumGenerations", "The maximum number of generations which should be processed."));
      Parameters.Add(new ValueLookupParameter<VariableCollection>("Results", "The variable collection where results should be stored."));
      Parameters.Add(new ValueLookupParameter<IOperator>("Analyzer", "The operator used to analyze each generation."));
      Parameters.Add(new ValueLookupParameter<IntValue>("EvaluatedSolutions", "The number of times solutions have been evaluated."));
      Parameters.Add(new ValueLookupParameter<IntValue>("PopulationSize", "The size of the population."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MinimumPopulationSize", "The minimum size of the population of solutions."));
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumPopulationSize", "The maximum size of the population of solutions."));
      Parameters.Add(new ValueLookupParameter<DoubleValue>("ComparisonFactor", "The comparison factor."));
      Parameters.Add(new ValueLookupParameter<IntValue>("Effort", "The maximum number of offspring created in each generation."));
      Parameters.Add(new ScopeParameter("CurrentScope", "The current scope which represents a population of solutions on which the genetic algorithm should be applied."));
      #endregion

      #region Create operators
      VariableCreator variableCreator = new VariableCreator();
      Assigner assigner1 = new Assigner();
      ResultsCollector resultsCollector1 = new ResultsCollector();
      Placeholder analyzer1 = new Placeholder();
      Placeholder selector = new Placeholder();
      SubScopesProcessor subScopesProcessor1 = new SubScopesProcessor();
      ChildrenCreator childrenCreator = new ChildrenCreator();
      UniformSubScopesProcessor uniformSubScopesProcessor1 = new UniformSubScopesProcessor();
      Placeholder crossover = new Placeholder();
      StochasticBranch stochasticBranch = new StochasticBranch();
      Placeholder mutator = new Placeholder();
      UniformSubScopesProcessor uniformSubScopesProcessor2 = new UniformSubScopesProcessor();
      Placeholder evaluator = new Placeholder();
      WeightedParentsQualityComparator comparator1 = new WeightedParentsQualityComparator();
      ConditionalSelector conditionalSelector = new ConditionalSelector();
      RightReducer rightReducer1 = new RightReducer();
      UniformSubScopesProcessor uniformSubScopesProcessor3 = new UniformSubScopesProcessor();
      SubScopesRemover subScopesRemover = new SubScopesRemover();
      DuplicateSelector duplicateSelector = new DuplicateSelector();
      LeftReducer leftReducer = new LeftReducer();
      SubScopesCounter subScopesCounter1 = new SubScopesCounter();
      IntCounter intCounter1 = new IntCounter();
      Comparator comparator2 = new Comparator();
      ConditionalBranch conditionalBranch1 = new ConditionalBranch();
      BestSelector bestSelector1 = new BestSelector();
      BestSelector bestSelector2 = new BestSelector();
      RightReducer rightReducer2 = new RightReducer();
      SubScopesProcessor subScopesProcessor2 = new SubScopesProcessor();
      BestSelector bestSelector3 = new BestSelector();
      RightReducer rightReducer3 = new RightReducer();
      MergingReducer mergingReducer = new MergingReducer();
      IntCounter intCounter2 = new IntCounter();
      Comparator comparator3 = new Comparator();
      Placeholder analyzer2 = new Placeholder();
      ConditionalBranch conditionalBranch2 = new ConditionalBranch();
      Assigner assigner2 = new Assigner();
      SubScopesCounter subScopesCounter2 = new SubScopesCounter();
      ResultsCollector resultsCollector2 = new ResultsCollector();
      Comparator comparator4 = new Comparator();
      ConditionalBranch conditionalBranch3 = new ConditionalBranch();

      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("Generations", new IntValue(0))); // Class GeneticAlgorithm expects this to be called Generations
      variableCreator.CollectedValues.Add(new ValueParameter<IntValue>("CurrentPopulationSize", new IntValue(0)));

      assigner1.Name = "Initialize CurrentPopulationSize";
      assigner1.LeftSideParameter.ActualName = "CurrentPopulationSize";
      assigner1.RightSideParameter.ActualName = EvaluatedSolutionsParameter.ActualName;

      resultsCollector1.CollectedValues.Add(new LookupParameter<IntValue>("Generations"));
      resultsCollector1.CollectedValues.Add(new LookupParameter<IntValue>("CurrentPopulationSize"));
      resultsCollector1.ResultsParameter.ActualName = "Results";

      analyzer1.Name = "Analyzer";
      analyzer1.OperatorParameter.ActualName = "Analyzer";

      selector.Name = "Selector";
      selector.OperatorParameter.ActualName = "Selector";

      childrenCreator.ParentsPerChild = new IntValue(2);

      crossover.Name = "Crossover";
      crossover.OperatorParameter.ActualName = "Crossover";

      stochasticBranch.ProbabilityParameter.ActualName = "MutationProbability";
      stochasticBranch.RandomParameter.ActualName = "Random";

      mutator.Name = "Mutator";
      mutator.OperatorParameter.ActualName = "Mutator";

      uniformSubScopesProcessor2.Parallel.Value = true;

      evaluator.Name = "Evaluator";
      evaluator.OperatorParameter.ActualName = "Evaluator";

      comparator1.ComparisonFactorParameter.ActualName = ComparisonFactorParameter.Name;
      comparator1.LeftSideParameter.ActualName = QualityParameter.Name;
      comparator1.MaximizationParameter.ActualName = MaximizationParameter.Name;
      comparator1.RightSideParameter.ActualName = QualityParameter.Name;
      comparator1.ResultParameter.ActualName = "SuccessfulOffspring";

      conditionalSelector.ConditionParameter.ActualName = "SuccessfulOffspring";
      conditionalSelector.ConditionParameter.Depth = 1;
      conditionalSelector.CopySelected.Value = false;

      duplicateSelector.CopySelected.Value = false;

      uniformSubScopesProcessor3.Parallel.Value = true;

      subScopesRemover.RemoveAllSubScopes = true;

      subScopesCounter1.Name = "Count Successful Offspring";
      subScopesCounter1.ValueParameter.ActualName = "NumberOfSuccessfulOffspring";

      intCounter1.IncrementParameter.ActualName = "NumberOfSuccessfulOffspring";
      intCounter1.IncrementParameter.Value = null;
      intCounter1.ValueParameter.ActualName = EvaluatedSolutionsParameter.Name;

      comparator2.Comparison = new Comparison(ComparisonType.GreaterOrEqual);
      comparator2.LeftSideParameter.ActualName = "NumberOfSuccessfulOffspring";
      comparator2.RightSideParameter.ActualName = MaximumPopulationSizeParameter.Name;
      comparator2.ResultParameter.ActualName = "SelectMaximum";

      conditionalBranch1.ConditionParameter.ActualName = "SelectMaximum";

      bestSelector1.CopySelected = new BoolValue(true);
      bestSelector1.MaximizationParameter.ActualName = MaximizationParameter.Name;
      bestSelector1.NumberOfSelectedSubScopesParameter.ActualName = MaximumPopulationSizeParameter.Name;
      bestSelector1.QualityParameter.ActualName = QualityParameter.Name;

      bestSelector2.CopySelected = new BoolValue(true);
      bestSelector2.MaximizationParameter.ActualName = MaximizationParameter.Name;
      bestSelector2.NumberOfSelectedSubScopesParameter.ActualName = "NumberOfSuccessfulOffspring";
      bestSelector2.QualityParameter.ActualName = QualityParameter.Name;

      bestSelector3.CopySelected = new BoolValue(false);
      bestSelector3.MaximizationParameter.ActualName = MaximizationParameter.Name;
      bestSelector3.NumberOfSelectedSubScopesParameter.ActualName = "Elites";
      bestSelector3.QualityParameter.ActualName = QualityParameter.Name;

      intCounter2.Increment = new IntValue(1);
      intCounter2.ValueParameter.ActualName = "Generations";

      comparator3.Comparison = new Comparison(ComparisonType.GreaterOrEqual);
      comparator3.LeftSideParameter.ActualName = "Generations";
      comparator3.ResultParameter.ActualName = "Terminate";
      comparator3.RightSideParameter.ActualName = MaximumGenerationsParameter.Name;

      analyzer2.Name = "Analyzer";
      analyzer2.OperatorParameter.ActualName = "Analyzer";

      conditionalBranch2.ConditionParameter.ActualName = "Terminate";

      assigner2.Name = "Reset CurrentPopulationSize";
      assigner2.LeftSideParameter.ActualName = "CurrentPopulationSize";
      assigner2.RightSideParameter.Value = new IntValue(0);

      subScopesCounter2.Name = "Set CurrentPopulationSize";
      subScopesCounter2.ValueParameter.ActualName = "CurrentPopulationSize";

      resultsCollector2.CollectedValues.Add(new LookupParameter<IntValue>("CurrentPopulationSize"));
      resultsCollector2.ResultsParameter.ActualName = "Results";

      comparator4.Comparison = new Comparison(ComparisonType.Less);
      comparator4.LeftSideParameter.ActualName = "CurrentPopulationSize";
      comparator4.RightSideParameter.ActualName = MinimumPopulationSizeParameter.Name;
      comparator4.ResultParameter.ActualName = "Terminate";

      conditionalBranch3.ConditionParameter.ActualName = "Terminate";
      #endregion

      #region Create operator graph
      OperatorGraph.InitialOperator = variableCreator;
      variableCreator.Successor = assigner1;
      assigner1.Successor = resultsCollector1;
      resultsCollector1.Successor = analyzer1;
      analyzer1.Successor = selector;
      selector.Successor = subScopesProcessor1;
      subScopesProcessor1.Operators.Add(new EmptyOperator());
      subScopesProcessor1.Operators.Add(childrenCreator);
      subScopesProcessor1.Successor = subScopesProcessor2;
      childrenCreator.Successor = uniformSubScopesProcessor1;
      uniformSubScopesProcessor1.Operator = crossover;
      uniformSubScopesProcessor1.Successor = uniformSubScopesProcessor2;
      crossover.Successor = stochasticBranch;
      stochasticBranch.FirstBranch = mutator;
      stochasticBranch.SecondBranch = null;
      stochasticBranch.Successor = null;
      mutator.Successor = null;
      subScopesRemover.Successor = null;
      uniformSubScopesProcessor2.Operator = evaluator;
      uniformSubScopesProcessor2.Successor = conditionalSelector;
      evaluator.Successor = comparator1;
      conditionalSelector.Successor = rightReducer1;
      rightReducer1.Successor = uniformSubScopesProcessor3;
      uniformSubScopesProcessor3.Operator = subScopesRemover;
      uniformSubScopesProcessor3.Successor = duplicateSelector;
      duplicateSelector.Successor = leftReducer;
      leftReducer.Successor = subScopesCounter1;
      subScopesCounter1.Successor = intCounter1;
      intCounter1.Successor = comparator2;
      comparator2.Successor = conditionalBranch1;
      conditionalBranch1.TrueBranch = bestSelector1;
      conditionalBranch1.FalseBranch = bestSelector2;
      bestSelector1.Successor = rightReducer2;
      bestSelector2.Successor = rightReducer2;
      subScopesProcessor2.Operators.Add(bestSelector3);
      subScopesProcessor2.Operators.Add(new EmptyOperator());
      subScopesProcessor2.Successor = mergingReducer;
      bestSelector3.Successor = rightReducer3;
      rightReducer3.Successor = null;
      mergingReducer.Successor = intCounter2;
      intCounter2.Successor = comparator3;
      comparator3.Successor = analyzer2;
      analyzer2.Successor = conditionalBranch2;
      conditionalBranch2.FalseBranch = assigner2;
      conditionalBranch2.TrueBranch = null;
      conditionalBranch2.Successor = null;
      assigner2.Successor = subScopesCounter2;
      subScopesCounter2.Successor = resultsCollector2;
      resultsCollector2.Successor = comparator4;
      comparator4.Successor = conditionalBranch3;
      conditionalBranch3.FalseBranch = selector;
      conditionalBranch3.TrueBranch = null;
      conditionalBranch3.Successor = null;
      #endregion
    }

    public override IOperation Apply() {
      if (CrossoverParameter.ActualName == null)
        return null;
      return base.Apply();
    }
  }
}
